#if ULTRA == 1
ROMBASE SET $4000
BANK SET 0
#else
ROMBASE SET $8000
#endif

        processor 6502
	org $04de

	; memory map:
	; $0300-$03ff sector buffer 
        ; $0400-$04ff flash buffer
	; $04de-$04ff startup code, overwritten later with data
	; $0500-$05ff main code

	;set up page size and source address
start   lda target
	sta $09
	lda #0
	sta $08
	lda track
	sta $0300
	lda sector
	sta $0301

	;read first sector of image file
	jsr rdsect
	bcs exit

	ldy #$00
	ldx binflg
	jmp load

	jmp start

target  dc.b $80	;target page in flash
bank	dc.b 0		;auto bank select, needs extra hardware
track   dc.b 17
sector  dc.b 0
binflg  dc.b 0		;0=bin, 2=prg
mirflg  dc.b $00	;mirror 8/16kB ROMs to $8000 ($80:enable, $00:disable)
pgsize  dc.b $80	;$40: 29C256, $80: 29C512,29C010, $00: 29C020,29C040
delay   dc.b 16		;16=2MHz, 8=1MHz

#if ULTRA == 1
endpg	dc.b $bf        ;end flag
#else
endpg	dc.b $ff        ;end flag
#endif

endflg  dc.b $fe	;254 bytes per sector
page    dc.b $00        ;current page
banktab	dc.b $00,$20,$40,$60


exit	sec
	rts

	;load flash buffer and call flash program
load	lda $0302,x
	sta $0400,y
	iny 
	beq pgm
ret	inx
	cpx endflg
	bne load
	jsr rdsect
	ldx #$00
	bcc load
end	rts

pgm	jsr pgmpage
	lda $09
	cmp endpg

#if BANK == 0
	beq end
#else
	bne ret2
	bit mirflg
	bvc end
	lda bank
	cmp #3
	beq end
	inc bank
	lda target
	sta $09
	bne ret3
#endif

ret2	inc $09
ret3	ldy #0
	beq ret

rdsect	lda $0301
	sta $07
	lda $0300
	sta $06
	beq exit

#if RETRY == 1		
	ldx #5		;# of retry
#endif
read	lda #$80
	sta $00
wait	lda $00		;wait for drive job to finish
	bmi wait
	cmp #2
#if RETRY == 1
	bcc okay
	dex
	bne read

exiterr	pla
	pla
	pla
	pla
	lda $0
	ldx #0
	jmp $e60a
#else
	bcs exiterr     
#endif

okay	lda $0300
	bne exit2
	ldx $0301
	dex
	stx endflg
exit2	clc
	rts

#if RETRY == 0
exiterr pla
	pla
	pla
	pla
	lda $0
	ldx #0
	jmp $e60a
#endif

	;load 256 bytes to flash in pages sized 64, 128 or 256 bytes
pgmpage	sei
	lda $1c00	; flash LED and set banking bits
	eor #8		; (optional connection of VIA lines to 
	pha		;  banking switches)
	and #%10011111
	ldy bank
	ora banktab,y
	sta $1c00

	ldy #0
nextpg	lda page
	clc
	adc pgsize
	sta page
	jsr wpflash
source	lda $0400,y
	sta ($08),y
	iny	      
	cpy page
	bne source

#if ULTRA ==1
waitprg lda $4000
	cmp $4000
	bne waitprg
#else
	jsr waitprg
#endif

	lda page
	bne nextpg

#if MIRROR == 1
	bit mirflg
	bpl out2
	lda $09
	cmp #$c0
	bcc out
	sec
	lda $09
	sbc #$40
	sta $09
	bne nextpg

out	clc
	lda $09
	adc #$40
	sta $09
#endif

out2	pla
	sta $1c00
	cli

#if ULTRA == 0	
	rts
#else
verify	ldy #0
vsource	lda $0400,y
	cmp ($08),y
	bne verror
	iny	      
	bne vsource
	rts
verror  lda #7
	sta $00
	jmp exiterr
#endif


#if ULTRA == 0
	;wait until current page has been flashed
waitprg txa
	pha
	tya
	pha
	ldx delay                ; write cycle takes at most 10ms
	ldy #0                   ; so we wait 16 x 256 x 5 x 0.5us
loop	iny                      ;(takes twice as long on 1541 1MHz)
	bne loop
	dex
	bne loop
	pla
	tay
	pla
	tax
	rts
#endif

	;turn off write protect of flash chip
wpflash	lda #$aa
	sta [$5555+ROMBASE]	;$5555
	lda #$55
	sta [$2aaa+ROMBASE]	;$2aaa
	lda #$a0
	sta [$5555+ROMBASE]	;$5555
	rts
